% Copyleft (c) 1997 by Tom Howland (tom@rahul.net)

:- module(xref, [xref/2]).

/* xref(+Files, +Options)

"Files" must be single prolog file specification or a ground list of
prolog file specifications. The ".pl" extension may be omitted.

"Options" must be a ground list of zero or more options of the form:

     verbose ... verbose output. This forces the cross referencer to
         echo the names of the files being read to user_error like qld
         would do.

     warnings(WFile) ...  warning file. Warnings are written to
         user_error by default. "WFile" must be a file specification.

     xref(XFile) ... cross reference file. The cross reference is
         written to user_output by default. "XFile" must be a file
         specification.

     mod(MFile) ... generate a file indicating which predicates are
         imported and which are exported for each file. This is not
         generated by default. "MFile" must be a file specification.

     undefined(UFile) ... generate a list of undefined predicates.
         References to undefined predicates are given in the warnings
         file, but sometimes you just want a list of the undefined
         predicates. "UFile" must be a file specification.

For example, the following would be a reasonable invocation:

	xref(foo, [verbose,xref('/dev/null')])

This would generate a cross reference of "foo.pl", the actual cross
reference listing is ignored (written to /dev/null), and the warnings
are displayed on user_error.

*/

:- use_module(library(fparse)).
:- use_module(library(avl)).
:- use_module(library(basics)).
:- use_module(library(strings)).
:- use_module(library(files)).

:- dynamic visited/1.		% so we won't visit a file twice.
:- dynamic cant_follow/7.	% a rare warning

xref(Files, Options) :-

    retractall(cant_follow(_,_,_,_,_,_,_)), retractall(visited(_)),

    default(compiler_errors, compiler, normal, Options, CompilerErrors),
    default(verbose, verbose, silent, Options, Verbose),
    default(warnings(WFile), WFile, user_error, Options, WFile),
    default(xref(XFile), XFile, user_output, Options, XFile),
    default(mod(MFile), MFile, suppress, Options, MFile),
    default(undefined(UFile), UFile, suppress, Options, UFile),

    current_output(CO), set_output(user_error),
    prolog_flag(gc_margin, Old, 9216),	% 9 times the default
    load_files(Files, [], user, '', 0-Verbose, d([],empty,[],[],empty), Dict),

    verbose(Verbose, thinking),
    retractall(visited(_)), garbage_collect_atoms,
    think(Dict, MFile, UFile, R, XREF, MMD),

    verbose(Verbose, printing),
    print_it(WFile, R, XFile, XREF, MFile, MMD, CompilerErrors),

    verbose(Verbose, done),
    prolog_flag(gc_margin, _, Old),
    set_output(CO).

default(Key, Value,     _, Options, Value) :- memberchk(Key, Options), !.
default(_,       _, Value,       _, Value).

verbose(verbose, M) :- !, write(M), nl, flush.
verbose(_, _).

flush :- current_output(CO), flush_output(CO).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/* this is where we read in the sources.

   The object is to create a d/5 structure representing the entire application.

   d(Defs, Imports, Externs, Initializations, MetaPredicates)

   where

   Defs .. a list of clauses, represented as Head-Body, where

       Head is a g/3 structure, that is, a term of the form

       		g(Name, Arity, Module)

       Body is a a term clause(File, LC, Goals) where

          File is the file in which the clause was read
	  LC is the line number
	  Goals is a list of g/3 structures.

   Imports .. an avl tree. g(N,A,OldModule)-g(N,A,NewModule)

   Externs .. a list of terms of the form

   	extern(File, LC, Name, Arity, Module)

   Initializations ... a list of terms of the form

   	initialization(File, LC, Goals)

      where Goals is a list of g/3 structures.

   MetaPredicates ... an avl tree g(Name,Arity,Module)-(Head-Metas) where

      Head is a term that will unify with a meta predicate
      Metas is a list of variables occuring in Head that are the meta arguments
*/

load_files([], _, _, _, _, D, D) :- !.
load_files([H|T], Options, M, F, L) --> !,
    load_files(H, Options, M, F, L), load_files(T, Options, M, F, L).
load_files(_, Options, _, _, _, D, D) :-
    memberchk(when(compile_time), Options),
    !.
load_files(X, Options, Mod, F, Level-Verbose, D0, D) :-
    quintus_file_name(X, Y), fparse(Y, F, Z),
    absolute_file_name(Z, [file_type(prolog),access(exist),file_errors(fail)],
		       File),
    !,
    open(File, read, Stream), read_one(Stream, Term, LC),
    (   Term = (:-module(M, Exports))
    ->  (   memberchk(imports(Imports), Options), Imports \== all
	->  true
	;   Imports = Exports
	),
	read_one(Stream, TheTerm, TheLC), TheMod = M, imports(Mod, M, Imports, D0, D1)
    ;   TheTerm = Term, TheLC = LC, TheMod = Mod, D1 = D0
    ),
    (   visited(File)
    ->  D = D1
    ;   (   Verbose == verbose
	->  format('~*c~a~n', [Level,32,File]), flush
	;   true
	),
	assert(visited(File)), L is Level+1,
        read_loop(TheTerm, Stream, TheLC, TheMod, File, L-Verbose, D1, D)
    ),
    close(Stream).
load_files(X, _, _, F, _, D, D) :-
    format('Can''t find the file ~q', [X]),
    (   F == ''
    ->  nl
    ;   format(' (referenced in the file ~a)~n', [F])
    ),
    flush.

quintus_file_name(Z, Atm) :- functor(Z, N, 1), !, user:file_search_path(N, X),
    quintus_file_name(X, A1), arg(1, Z, R), quintus_file_name(R, A2),
    concat_atom([A1,/,A2], Atm).
quintus_file_name(X, X).

read_one(Stream, X, LC) :- repeat,
    read_term(Stream, [term_position('$stream_position'(_,LC,_,_,_))], Y),
    !, expand_term(Y, X).

imports(Mod, M, Exports, d(DEFS,Imports0,Externs,INI,Mt),
	d(DEFS,Imports,Externs,INI,Mt)
       ) :- import(Exports, Mod, M, Imports0, Imports).

import([], _, _, I, I).
import([N/A|R], Mod, M, I0, I) :-
    avl_store(g(N,A,Mod), I0, g(N,A,M), I1), import(R, Mod, M, I1, I).

read_loop(end_of_file, _, _,   _,    _, _,  D, D) :- !.
read_loop(X0, Stream, LC0, Mod, File, L, D0, D) :-
    recognize(X0, LC0, Mod, File, L, D0, D1),
    read_one(Stream, X, LC),
    read_loop(X, Stream, LC, Mod, File, L, D1, D).

recognize(foreign_file(_,_), _, _, _, _, D, D) :- !.
recognize([], _,  _, _, _, D, D) :- !.
recognize([H|T], LC, Mod, File, L) --> !,
    recognize(H, LC, Mod, File, L), recognize(T, LC, Mod, File, L).
recognize((:-G), LC, Mod, File, L) --> !, command(G, LC, Mod, File, L).
recognize((M:X), LC, _, File, L) --> !, recognize(X, LC, M, File, L).
recognize((X:-Y), LC, Mod, File, _, D0, D) :- !,
    module_prefix(X, Mod, Head, HeadC),
    insert_clause(Head, Y, HeadC, LC, Mod, File, D0, D).
recognize(foreign(_,_,G), LC, Mod, File, _) --> !,
    insert_clause(G, true, Mod, LC, Mod, File).
recognize(foreign(_,G), LC, Mod, File, _) --> !,
    insert_clause(G, true, Mod, LC, Mod, File).
recognize(X, LC, Mod, File, _) --> insert_clause(X, true, Mod, LC, Mod, File).

command((X,Y), LC, M, F, L) --> !,
    command(X, LC, M, F, L), command(Y, LC, M, F, L).
command(extern(G), _, _, _, _, D, D) :- predicate_property(G, built_in), !.
command(extern(G), LC, Mod, File, _, d(D,I,E,N,Mt),
	d(D,I,[extern(File,LC,N1,A1,Mod)|E],N,Mt)) :- !,
    functor(G, N1, A1).
command(initialization(Body), LC, Mod, File, _, d(D,I,E,N,Mt),
	d(D,I,E,[initialization(File,LC,B)|N],Mt)) :- !,
    body(Body, x('', initialization,1,LC,Mod,File,I,Mt), B, []).
command(meta_predicate(Meta), _, Mod, _, _, d(D,I,E,N,M0), d(D,I,E,N,M)) :- !,
    parse_metas(Meta, Mod, M0, M).
command(dynamic(Dynamic), LC, Mod, File, _) --> !,
    parse_dynamics(Dynamic, LC, Mod, File).
command(LoadCommand, _, M, F, L, D0, D) :-
    load_command(LoadCommand, Files, Options),
    !,
    load_files(Files, Options, M, F, L, D0, D).
command(_, _, _, _, _, D, D).

parse_metas(M:X, _) --> !, parse_metas(X, M).
parse_metas((X,Y), Mod) --> !, parse_metas(X, Mod), parse_metas(Y, Mod).
parse_metas(Meta, Mod, M0, M) :- functor(Meta, N, A), functor(X, N, A),
    parse_meta(A, Meta, X, L), avl_store(g(N,A,Mod), M0, X-L, M).

parse_meta(0, _, _, []) :- !.
parse_meta(A, Meta, X, L) :-
    (   arg(A, Meta, Arg),
        (   Arg == :
        ->  true
	;   integer(Arg)
	)
    ->  L = [H|T], arg(A, X, H)		% the crux of the biscuit
    ;   L = T
    ),
    B is A-1, parse_meta(B, Meta, X, T).

parse_dynamics((X,Y), LC, Mod, File) --> !,
    parse_dynamics(X, LC, Mod, File), parse_dynamics(Y, LC, Mod, File).
parse_dynamics(M:N/A, LC, _, File) --> !, dynamic(N, A, LC, M, File).
parse_dynamics(N/A, LC, Mod, File) --> !, dynamic(N, A, LC, Mod, File).

dynamic(N, A, LC, Mod, File) --> insert_clause(N, A, true, Mod, LC, Mod, File).

load_command(load_files(Files, Options), Files, Options).
load_command(load_files(Files), Files, []).
load_command([H|T], [H|T], []).
load_command(compile(Files), Files, [load_type(compile)]).
load_command(consult(Files), Files, [load_type(compile)]).
load_command(ensure_loaded(Files), Files, [if(changed)]).
load_command(use_module(Files), Files, [must_be_module(yes)]).
load_command(use_module(Files,Imports), Files,
	     [must_be_module(yes),imports(Imports)]).
load_command(use_module(_, File, Imports), File,
	     [must_be_module(yes),imports(Imports)]) :- ground(File).

module_prefix(M:X, _, Head, HeadC) :- !, module_prefix(X, M, Head, HeadC).
module_prefix(Head, Mod, Head, Mod).

insert_clause(N, A, Y, HeadC, LC, BodyC, File, D0, D) :-
    functor(X, N, A), insert_clause(X, Y, HeadC, LC, BodyC, File, D0, D).

insert_clause(X, Y, HeadC, LC, BodyC, File, d(T,I,E,IM,Mt), d([H|T],I,E,IM,Mt))
:-
    H = G-clause(File,LC,Body),
    G = g(N,A,HeadC),
    functor(X, N, A),
    body(Y, x(G,N,A,LC,BodyC,File,I,Mt), Body, []).

body(X, M, S, S) :- var(X), !, cant_follow(M).
body((M:X), MM)                  --> !, change_mod(X, M, MM).
body((\+X), M)                   --> !, body(X, M).
body(abolish(X), M)              --> !, pred_spec_tree(abolish, 2, X, M).
body(assert(X), M)               --> !, parent(assert, 1, X, M).
body(asserta(X), M)              --> !, parent(asserta, 1, X, M).
body(assertz(X), M)              --> !, parent(assertz, 1, X, M).
body(call(X), M)                 --> !, parent(call, 1, X, M).
body(multifile_assertz(X), M)    --> !, parent(multifile_assertz, 1, X, M).
body(retract(X), M)              --> !, parent(retract, 1, X, M).
body(retractall(X), M)           --> !, parent(retractall, 1, X, M).
body((X,Y), M)                   --> !, body(X, M), body(Y, M).
body((X->Y), M)                  --> !, body(X, M), body(Y, M).
body((X;Y), M)                   --> !, body(X, M), body(Y, M).
body(_^X, M)                     --> !, body(X, M).
body(abolish(N,A), M)            --> !, pred_spec_tree(abolish, 2, N/A, M).
body(assert(X,_), M)             --> !, parent(assert, 2, X, M).
body(asserta(X,_), M)            --> !, parent(asserta, 2, X, M).
body(assertz(X,_), M)            --> !, parent(assertz, 2, X, M).
body(clause(X,_), M)             --> !, parent(clause, 2, X, M).
body(current_predicate(_,X), M)  --> !, parent(current_predicate, 2, X, M).
body(phrase(X,_), M)             --> !, phrase(phrase, 2, X, M).
body(predicate_property(X,_), M) --> !, parent(predicate_property, 2, X, M).
body(save_predicates(X,_), M)	 --> !, pred_spec_tree(save_predicates, 2, X, M).
body(source_file(X,_), M)        --> !, parent(source_file, 2, X, M).
body(add_advice(X,_,Y), M0)      --> !, {parent(add_advice, 3, M0, M)},
    body(X, M), body(Y, M).
body(bagof(_,X,_), M)            --> !, parent(bagof, 3, X, M).
body(clause(X,_,_), M)           --> !, parent(clause, 3, X, M).
body(current_advice(X,_,Y), M0)  --> !, {parent(current_advice, 3, M0, M)},
    body(X, M), body(Y, M).
body(findall(_,X,_), M)          --> !, parent(findall, 3, X, M).
body(on_exception(_,X,Y), M0)    --> !, {parent(on_exception, 3, M0, M)},
    body(X, M), body(Y, M).
body(phrase(X,_,_), M)           --> !, phrase(phrase, 3, X, M).
body(setof(_,X,_), M)            --> !, parent(setof, 3, X, M).
body(source_file(X,_,_), M)      --> !, parent(source_file, 3, X, M).
body(X, _, S, S) :- predicate_property(X, built_in), !.
body(X, MM, [G|S0], S) :- functor(X, N, A), arg(5, MM, M),
    maybe_meta(X, MM, g(N,A,M), G, S0, S).

change_mod(_, M, X, S, S) :- var(M), !, cant_follow(X).
change_mod(X, M, x(G,N,A,LC,_,F,I,Mt)) --> body(X, x(G,N,A,LC,M,F,I,Mt)).

cant_follow(x(G,N,A,LC,_,F,_,Mt)) :-
   (  functor(NF, N, A), need_not_follow(NF)
   -> true
   ;  avl_fetch(G, Mt)
   -> true
   ;  g(Ng,Ag,Mg) = G, assert(cant_follow(N,A,Mg,Ng,Ag,F,LC))
   ).

need_not_follow(assert(_,_)).
need_not_follow(assert(_)).
need_not_follow(abolish(_)).
need_not_follow(abolish(_,_)).

parent(N, A, X, M0, S0, S) :- parent(N, A, M0, M), body(X, M, S0, S).

parent(N, A, x(G,_,_,LC,M,F,I,Mt), x(G,N,A,LC,M,F,I,Mt)).

pred_spec_tree(N, A, Tree, M0, S0, S) :- parent(N, A, M0, M),
    pred_spec_tree(Tree, M, S0, S).

pred_spec_tree([], _, S, S) :- !.
pred_spec_tree([H|T], M) --> !, pred_spec_tree(H, M), pred_spec_tree(T, M).
pred_spec_tree(M:_, MM, S, S) :- var(M), !, cant_follow(MM).
pred_spec_tree(M:X, x(G,N,A,LC,_,F,I,Mt)) --> !,
    pred_spec_tree(X, x(G,N,A,LC,M,F,I,Mt)).
pred_spec_tree(N/A, M, S, S) :- \+ (atom(N), integer(A)), !, cant_follow(M).
pred_spec_tree(N/A, M, S0, S) :- functor(X, N, A), body(X, M, S0, S).

phrase(Phrase, Ap, X, M0, S0, S) :-
    (  var(X)
    -> parent(Phrase, Ap, M0, M), S0 = S, cant_follow(M)
    ;  expand_term((x-->X),(_:-Y)), body(Y, M0, S0, S)
    ).

maybe_meta(X, M0, G, G, S0, S) :-
    arg(8, M0, Metas),
    avl_fetch(G, Metas, V),
    !,
    copy_term(V, X-L),
    functor(X, N, A),
    parent(N, A, M0, M),
    meta(L, M, S0, S).
maybe_meta(X, M, G0, G, S0, S) :-
    arg(7, M, Imports),
    avl_fetch(G0, Imports, G1),
    !,
    maybe_meta(X, M, G1, G, S0, S).
maybe_meta(_, _, G, G, S, S).

meta([], _, S, S).
meta([H|T], M) --> body(H, M), meta(T, M).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
% The list of clauses is converted into an avl tree.  The cross
% referencing is then done by following runtime_entry/1, externs, and
% initializations. Finally, the results are prepared for printing.

think(d(Defs,Imports,Externs,INI,_), MFile, UFile, Warnings, XREF, MMD) :-
    keysort(Defs, Sorted),
    merge_clauses(Sorted, Merged),
    ord_list_to_avl(Merged, Def0),
    follow_top(Imports, Externs, INI, Def0, Def),
    massage(Def, MFile, UFile, Warnings, XREF, MMD).

follow_top(Imports, Externs, INI) -->
    follow_initializations(INI, Imports),
    follow(g(runtime_entry,1,user), xref, Imports),
    follow_hooks([message_hook/3,generate_message_hook/3,portray/1,
		  query_hook/6,unknown_predicate_handler/3,user_help/0,
		  term_expansion/2],
		user, Imports),
    follow_hooks([generate_message/3,query_abbreviation/3], 'QU_messages',
		 Imports),
    follow_externs(Externs, Imports).

merge_clauses([], []).
merge_clauses([K-H|T0], [K-clauses([H|L])|R]) :-
    merge(T0, K, L, T), merge_clauses(T, R).

merge([K-H|T], K, [H|R], ML) :- !, merge(T, K, R, ML).
merge(T, _, [], T).

follow_initializations([], _, Def, Def).
follow_initializations([initialization(F,LC,B)|T], Imports) -->
    follow_body(B, initialization(F,LC), Imports),
    follow_initializations(T, Imports).

follow_body([], _, _, Def, Def).
follow_body([H|T], Referee, Imports) -->
    follow(H, Referee, Imports), follow_body(T, Referee, Imports).

follow(G, Referee, Imports, Def0, Def) :-
    (   avl_fetch(G, Def0, V)
    ->  found(V, G, Referee, Imports, Def0, Def)
    ;   avl_fetch(G, Imports, K)
    ->  follow(K, Referee, Imports, Def0, Def)
    ;   (   Referee == hook
	->  true
	;   undefined_ok(G)
	)
    ->  Def0 = Def
    ;   avl_store(G, Def0, undefined([Referee]), Def)
    ).

found(clauses(Clauses), G, Referee, Imports, Def0, Def) :-
    collect_files(Clauses, Files1), to_ordset(Files1, Files, []),
    avl_change(G, Def0, _, Def1, referenced(Files,[Referee])),
    g(N,A,M) = G, follow_bodies(Clauses, N, A, M, Imports, Def1, Def).
found(referenced(Files,R), G, Referee, _, Def0, Def) :-
    avl_change(G, Def0, _, Def, referenced(Files,[Referee|R])).
found(undefined(R), G, Referee, _, Def0, Def) :-
    avl_change(G, Def0, _, Def, undefined([Referee|R])).

collect_files([], _).
collect_files([clause(File,_,_)|T], Files) :-
    insert(File, Files), collect_files(T, Files).

insert(Name, dic(N,B,A)) :-
    (   Name = N
    ->  true
    ;   Name @< N
    ->  insert(Name, B)
    ;   insert(Name, A)
    ).

to_ordset(X, S, S) :- var(X), !.
to_ordset(dic(N,B,A)) --> to_ordset(B), [N], to_ordset(A).

follow_bodies([], _, _, _, _, Def, Def).
follow_bodies([clause(File,LC,Body)|T], N, A, M, Imports) -->
    follow_body(Body, referee(File,LC,N,A,M), Imports),
    follow_bodies(T, N, A, M, Imports).

undefined_ok(g(N,A,M)) :- undefined_ok(N, A, M).

undefined_ok(file_search_path, 2, user).
undefined_ok(library_directory, 1, user).
undefined_ok(parse_message, 3, 'QU_messages').
undefined_ok(parse_message_hook, 3, user).

follow_hooks([], _, _, Def, Def).
follow_hooks([N/A|T], M, Imports) -->
    follow(g(N,A,M), hook, Imports), follow_hooks(T, M, Imports).

follow_externs([], _, Def, Def).
follow_externs([extern(F,LC,N,A,M)|T], Imports) -->
    follow(g(N,A,M), extern(F,LC), Imports), follow_externs(T, Imports).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get the data ready for output

massage(Def0, MFile, UFile, Warnings, XREF, MODULE_FINISH) :-
    avl_to_list(Def0, L0),
    unreachable_files(L0, Def0, Def, empty, UR1),
    avl_to_list(UR1, UR2),
    unreachable_f(UR2, Warnings, W),
    avl_to_list(Def, L),
    wx(L, UR1, UnsortedWarnings0, X),
    cnt_follow(UnsortedWarnings0, UnsortedWarnings),
    sort(UnsortedWarnings, W),
    sort(X, XREF),
    (   MFile == suppress
    ->  true
    ;   module_create(L, empty, MODULE_CREATE),
        module_finish(MODULE_CREATE, MODULE_FINISH)
    ),
    maybe_collect_undefined(UFile, L).

% collect unreachable files and unreachable references to undefined predicates

unreachable_files([], Def, Def, D, D).
unreachable_files([K-V|T], D0, D) -->
    urf(V, K, D0, D1), unreachable_files(T, D1, D).

urf(clauses(Clauses), K, D0, D) --> urfc(Clauses, K, D0, D).
urf(referenced(Files,_), _, D, D) --> urfr(Files).
urf(undefined(_), _, Df, Df, D, D).

urfc([], _, Df, Df, D, D).
urfc([clause(File,LC,Body)|T], K, Df0, Df, D0, D) :-
    ur_body(Body, File, LC, K, Df0, Df1),
    urfcc(File, D0, D1),
    urfc(T, K, Df1, Df, D1, D).

ur_body([], _, _, _, D, D).
ur_body([H|T], File, LC, K, Def0, Def) :-
    (   avl_fetch(H, Def0, V)
    ->  ur_found(V, H, File, LC, K, Def0, Def1)
    ;   avl_store(H, Def0, undefined([not_called(File,LC,K)]), Def1)
    ),
    ur_body(T, File, LC, K, Def1, Def).

ur_found(clauses(_), _, _, _, _, D, D).
ur_found(referenced(_,_), _, _, _, _, D, D).
ur_found(undefined(U), H, File, LC, K, D0, D) :-
    avl_change(H, D0, _, D, undefined([not_called(File,LC,K)|U])).

urfcc(File, Dic, Dic) :- avl_fetch(File, Dic), !.
urfcc(File, Dic0, Dic) :- avl_store(File, Dic0, unreachable, Dic).

urfr([], D, D).
urfr([H|T], D0, D) :- avl_store(H, D0, reached, D1), urfr(T, D1, D).

unreachable_f([], S, S).
unreachable_f([F-K|T]) --> urfk(K, F), unreachable_f(T).

urfk(reached, _) --> [].
urfk(unreachable, F) --> [unreachable(F)].

% generate the cross reference list and the rest of the warnings
% wx ... warnings + cross reference

wx([], _, [], []).
wx([g(N,A,M)-V|T], UR, L0, X0) :- wx1(V, UR, N, A, M, L0, L, X0, X),
    wx(T, UR, L, X).

wx1(clauses([clause(F,LC,_)|_]), UR, N, A, M, [p(F,LC,not_reached(N,A,M))|L], L,
   X, X) :- avl_fetch(F, UR, reached), !.
wx1(clauses(_), _, _, _, _, L, L, X, X).
wx1(undefined(R), _, N, A, M, L0, L, X, X) :- undefined(R, N, A, M, L0, L).
wx1(referenced(Files, R), _, N, A, M, L, L, [ref(M,N,A,Files,RS)|X], X) :-
    referenced(R, R1), to_ordset(R1, RS, []).

undefined([], _, _, _, L, L).
undefined([H|T], N, A, M) --> undefined1(H, N, A, M), undefined(T, N, A, M).

undefined1(xref,                  _,  _,  _) -->
    [runtime_entry_warning].
undefined1(referee(F,LC,N,A,M),  N1, A1, M1) -->
    [p(F,LC,referenced(M1,N1,A1,M,N,A))].
undefined1(initialization(F,LC), N,   A,  M) -->
    [p(F,LC,initialization(N,A,M))].
undefined1(extern(F,LC),         N,   A,  M) -->
    [p(F,LC,extern(N,A,M))].
undefined1(not_called(F,LC,K), N, A, M) --> [p(F,LC,not_called(M,N,A,K))].

referenced([], _).
referenced([H|T], D) :- referenced1(H, D), referenced(T, D).

referenced1(xref, _). 
referenced1(referee(_,_,N,A,M),   D) :- insert(referee(M,N,A),       D).
referenced1(initialization(F,LC), D) :- insert(initialization(F,LC), D).
referenced1(extern(F,LC),         D) :- insert(extern(F,LC),         D).
referenced1(hook,                 D) :- insert(hook,                 D).

cnt_follow(L0, L) :-
    findall(p(F,LC,cant_follow(N,A,Mg,Ng,Ag)),
	    retract(cant_follow(N,A,Mg,Ng,Ag,F,LC)), X),
    append(L0, X, L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% create the data structures for module file output

module_create([], D, D).
module_create([G-H|T], D0, D) :- module_create1(H, G, D0, D1),
    module_create(T, D1, D).

module_create1(clauses(_),        _, D,  D).
module_create1(undefined(_),      _, D,  D).
module_create1(referenced([F],R), G, D0, D) :- !, % don't know how to do
    module_create2(R, F, G, D0, D).		  % multifile yet
module_create1(referenced(_,_),   _, D,  D).

module_create2([], _, _, D, D).
module_create2([H|T], F, G, D0, D) :- module_create3(H, F, G, D0, D1),
    module_create2(T, F, G, D1, D).

module_create3(xref,                 _, _, D,  D).
module_create3(hook,                 _, _, D,  D).
module_create3(initialization(F,_),  F, _, D,  D) :- !.
module_create3(initialization(FI,_), F, G, D0, D) :- ei(FI, F, G, D0, D).
module_create3(extern(F,_),          F, _, D,  D) :- !.
module_create3(extern(FI,_),         F, G, D0, D) :- ei(FI, F, G, D0, D).
module_create3(referee(F,_,_,_,_),   F, _, D,  D) :- !.
module_create3(referee(FI,_,_,_,_),  F, G, D0, D) :- ei(FI, F, G, D0, D).

% ei ... exports imports. This is fun. We have an avl tree whose keys
% are file names and whose values are pairs. The left member of a pair
% is an absorptive key tree indicating the predicates exported, the
% right member is an absorptive key-value tree, the keys of which are
% imported file names, and the values of which are absorptive key trees
% indicating the predicates imported. The terminology is my own. Sorry.

ei(ImportingFile, ExportingFile, g(Name,Arity,_), D0, D) :-
    Goal = [Name|Arity],
    fetchm(ExportingFile, D0, Exports, _, D1),
    insert(Goal, Exports),
    fetchm(ImportingFile, D1, _, ImportedFiles, D),
    insert(ExportingFile, ImportedFiles, Imports),
    insert(Goal, Imports).

fetchm(K,  A, E, I, A) :- avl_fetch(K,  A, E-I), !.
fetchm(K, A0, E, I, A) :- avl_store(K, A0, E-I, A).

insert(Name, dic(N,V,B,A), Value) :-
    (   Name = N
    ->  Value = V
    ;   Name @< N
    ->  insert(Name, B, Value)
    ;   insert(Name, A, Value)
    ).

module_finish(D, L) :- avl_to_list(D, L1), module_finish1(L1, L).

module_finish1([], []).
module_finish1([F-(E-I)|T], [x(F,Exports,Imports)|R]) :-
    to_ordset(E, Exports, []), m_imports(I, Imports, []), module_finish1(T, R).

m_imports(D, S, S) :- var(D), !.
m_imports(dic(N,V,B,A), S0, S) :-
    m_imports(B, S0, [N-X|S1]), to_ordset(V, X, []), m_imports(A, S1, S).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

maybe_collect_undefined(suppress, _) :- !.
maybe_collect_undefined(F, []) :- !,
    (   file_exists(F)
    ->  delete_file(F)
    ;   true
    ).
maybe_collect_undefined(UFile, L) :- collect_undefined(L, empty, U0),
    avl_to_list(U0, U), telling(Telling), tell(UFile),
    print_undefined(U), told, tell(Telling).

collect_undefined([], X, X).
collect_undefined([G-H|T], D0, D) :- cu(H, G, D0, D1),
    collect_undefined(T, D1, D).

cu(undefined(_), G, D0, D) :- !, avl_store(G, D0, foo, D).
cu(_, _, X, X).

print_undefined([]).
print_undefined([g(N,A,_)-_|T]) :- write(N), put(0'/), write(A), nl,
    print_undefined(T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% print it

print_it(WFile, Warnings, XFile, XREF, MFile, MODULE_FINISH, CompilerErrors) :-

    print_warnings(WFile, Warnings, CompilerErrors),
    print_cross_references(XFile, XREF),
    print_modules(MFile, MODULE_FINISH).

% print the warnings

print_warnings(user_error, Warnings, CE) :- !,
    print_warnings_top(Warnings, CE), flush_output(user_error).
print_warnings(File, Warnings, CE) :- telling(T), tell(File),
    print_warnings_top(Warnings, CE), told, tell(T).

print_warnings_top(Warnings, CE) :-
    length(Warnings, L), write(L), write(' warnings'), nl, nl,
    print_warning_list(Warnings, foo(boo), CE).

print_warning_list([], _, _).
print_warning_list([P|T], Old, CE) :- print_warning(P, Old, F, CE), nl,
    print_warning_list(T, F, CE).

print_warning(runtime_entry_warning, O, O, _) :-
    write('runtime_entry/1 is not defined').
print_warning(unreachable(F), O, O, _) :-
    write('nothing in file '), write(F), write(' was called').
print_warning(p(F,LC,Ms), O, F, CE) :-
   (  CE == normal
   -> (  O == F
      -> true
      ;  nl, write(F), nl
      ),
      write('    '), write(LC)
   ;  format('~w:~w:', [F,LC])
   ),
   print_warning(Ms).

print_warning(not_reached(N,A,M)) :-
    write(' unreachable '), write_pred_spec(M, N, A).
print_warning(referenced(Mu,Nu,Au,M,N,A)) :-
    put(32), referenced_undefined(M, N, A, Mu, Nu, Au).
print_warning(initialization(N,A,M)) :-
    write(' initialization refers to undefined predicate '),
    write_pred_spec(M, N, A).
print_warning(extern(N,A,M)) :-
    write(' extern declaration for undefined predicate '),
    write_pred_spec(M, N, A).
print_warning(not_called(Mu,Nu,Au,g(N,A,M))) :-
    write(' unreachable '), referenced_undefined(M, N, A, Mu, Nu, Au).
print_warning(cant_follow(N,A,Mg,Ng,Ag)) :-
    write(' can not follow '), writeq(N), put(0'/), write(A), write(' in '),
    write_pred_spec(Mg, Ng, Ag).

write_pred_spec(M, N, A) :- writeq(M), put(0':), writeq(N), put(0'/), write(A).

referenced_undefined(M, N, A, Mu, Nu, Au) :-
    write_pred_spec(M, N, A), write(' referenced undefined '),
    write_pred_spec(Mu, Nu, Au).

% print the cross reference listing

print_cross_references(user_output, X) :- !,
    current_output(CO), set_output(user_output), nl,
    print_xref(X), flush_output(user_output), set_output(CO).
print_cross_references(XFile, X) :-
    telling(T), tell(XFile), print_xref(X), told, tell(T).

print_xref(X) :- length(X, L), format('~d referenced predicates~n', [L]),
    print_xref_list(X).

print_xref_list([]).
print_xref_list([ref(M,N,A,Files,References)|T]) :- nl,
    write_pred_spec(M, N, A), nl,
    print_files(Files), print_references(References), print_xref_list(T).

print_files([]).
print_files([H|T]) :- write('  '), write(H), nl, print_files(T).

print_references([]).
print_references([H|T]) :-
    write('    '), print_reference(H), nl, print_references(T).

print_reference(referee(M,N,A)) :- write_pred_spec(M, N, A).
print_reference(initialization(F,LC)) :- write(initialization), line_ref(F, LC).
print_reference(extern(F,LC)) :- write('extern declaration'), line_ref(F, LC).
print_reference(hook)         :- write(hook).

line_ref(F, LC) :- write(' in file '), write(F), write(' at line '), write(LC).

% print the imports & exports file

print_modules(suppress, _) :- !.
print_modules(MFile, X) :-
    telling(T), tell(MFile), print_module(X), told, tell(T).

print_module([]).
print_module([x(F,Exports,Imports)|T]) :-
    fparse(F, _, N, _),
    format('% ~a ~2n:- module(~a, [~n', [F,N]),
    print_module_exports(Exports, ''),
    format(']).~2n', []),
    print_module_imports(Imports),
    nl,
    print_module(T).

print_module_exports([], _).
print_module_exports([[N|A]|T], Suffix) :-
    format(Suffix, []),
    format('			~q', [N/A]),
    print_module_exports(T, ',~n').

print_module_imports([]).
print_module_imports([F-H|T]) :-
    format(':- use_module(~q, [~n', [F]),
    print_module_exports(H, ''),
    format(']).~2n', []),
    print_module_imports(T).

end_of_file.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- use_module(library(addportray)).
:- use_module(library(files)).

my_avl_portrayal(node(_,_,_,_,_)) :- write(avl).

:- add_portray(my_avl_portrayal).

av(N, A) :- functor(X, N, A), add_advice(X, fail, trace).

clean :- close_all_streams, retractall(visited(_)),
    (current_advice(X,Y,Z), remove_advice(X, Y, Z), fail ; true),
    av(load_files, 7), av(body, 4), av(think, 5), av(follow, 5),
    av(massage, 5), av(unreachable_files, 5), av(print_it, 6),
    check_advice.

t(X) :- xref(X, [verbose]).
tm(X) :- xref(X, [mod('/tmp/mod'),verbose]).
